from pathlib import Path
print('Running' if __name__ == '__main__' else 'Importing', Path(__file__).resolve())

##########################################################
# Bachelor's thesis                                      #
# From a dumb boiler to a smart one using a smart socket #
# Author: Adam Grünwald                                  #
# BUT FIT BRNO, Faculty of Information Technology        #
# 26/6/2021                                              #
#                                                        #
# Module with class representing boiler itself.          #
# Serves for computing time needed to heat to certain    #
# temperature or recalculates number of showers          #
# to needed temperature to heat to.                      #
##########################################################

from datetime import datetime, timedelta
import time

from numpy import ndarray
from smartboiler.switch import Switch
import pandas as pd


class Boiler(Switch):
    def __init__(self,  base_url, token, headers, switch_entity_id, capacity=100, wattage=2000, set_tmp=60, one_shower_volume=40, shower_temperature=40, min_tmp=37, heater_efficiency=0.98):

        print("------------------------------------------------------\n")
        print('initializing of control...\n\tCapacity of Boiler = {}\n\t Wattage of boiler = {}\n'.format(
            capacity, wattage))
        print("------------------------------------------------------\n")
        super.__init__(self, ntity_id=switch_entity_id, url=base_url, token=token, headers=headers)
        self.boiler_heat_cap = capacity * 1.163
        self.real_wattage = wattage * heater_efficiency
        
        self.set_tmp = set_tmp
        self.capacity = capacity
        self.one_shower_volume = one_shower_volume
        self.shower_temperature = shower_temperature
        self.min_tmp = min_tmp
        self.area_tmp = 13
        self.boiler_measured_max = 35.6

    def time_needed_to_heat_up_minutes(self, consumption_kWh):
        """
        Calculates the time needed for heating water in boiler by temperature delta.
            time = (m * c) * d'(tmp) / P * effectivity_coef

            https://vytapeni.tzb-info.cz/tabulky-a-vypocty/97-vypocet-doby-ohrevu-teple-vody
        """

        return (consumption_kWh) / (self.real_wattage)

    def is_needed_to_heat(self, tmp_act:int, prediction_of_consumption:ndarray):
        """Conciders if it is needed to heat.

        Args:
            tmp_act ([type]): [actual temperature of water in boiler]
            tmp_goal ([type]): [temperature of water to heat before consumption]
            time_to_consumption ([type]): [time to consumption]

        Returns:
            [boolean]: [boolean value of needed to heat]
        """
        if(tmp_act < self.min_tmp):
            return True
        
        # get whole dataframe and check if it is needed to heat
        # if it is, return True
        # if not, pop last row and continue with reduced df
        # if df is empty, return False
        len_of_df = len(prediction_of_consumption)
        for i in range(len_of_df, 0):
            sum_of_consumption = prediction_of_consumption.iloc[:i].sum()
            time_to_consumption_minutes = i*30
            
            time_needed_to_heat = self.time_needed_to_heat_up_minutes(consumption_kWh=sum_of_consumption)
            
            if(time_to_consumption_minutes < time_needed_to_heat):
                return True
            
        return False
            

    def showers_degrees(self, number_of_showers):
        """Recalculates number of showers to temperature 
        on which is needed to heat up water in boiler.

        Args:
            number_of_showers ([int]): [number of showers to prepare]

        Returns:
            [float]: [tempreture to heat on]
        """
        showers_volume = number_of_showers * self.one_shower_volume

        cold_water_tmp = 10

        needed_temperature = ((self.min_tmp * self.capacity) + (self.shower_temperature *
                              showers_volume) - (showers_volume * cold_water_tmp)) / self.capacity

        if needed_temperature > self.set_tmp:
            return self.set_tmp

        return needed_temperature

    def real_tmp(self, tmp_act):
        """Calculates the real temperature of water in boiler

        Args:
            tmp_act ([float]): [actual temperature measured in boiler]

        Returns:
            [float]: [the real temperature of water in boiler]
        """

        if((tmp_act is None) or (self.area_tmp is None) or (self.set_tmp is None)):
            return 50

        if(tmp_act < self.area_tmp or tmp_act > self.set_tmp):
            return tmp_act

        tmp_act_and_area_delta = tmp_act - self.area_tmp
        tmp_max_and_area_delta = self.boiler_measured_max - self.area_tmp

        p1 = tmp_act_and_area_delta / tmp_max_and_area_delta

        tmp = p1 * (self.set_tmp - self.area_tmp) + self.area_tmp
        return tmp

    def set_measured_tmp(self, df):
        """Initializes measured temperatures for calculation of real temperature of water in boiler.

        Args:
            df ([DataFrame]): [dataframe with measured data]
        """
        df_of_last_week = df[df.index > (
            df.last_valid_index() - timedelta(days=21))]

        self.area_tmp = df_of_last_week['tmp1'].nsmallest(100).mean()
        self.boiler_measured_max = df_of_last_week['tmp2'].nlargest(100).mean()

        print("area_tmp: ", self.area_tmp,
              "\nboiler_max: ", self.boiler_measured_max)


if __name__ == '__main__':
    tmp_act = 30
    tmp_goal = 52
    time_to_consumption = 1
    b = Boiler(capacity=80, wattage=2000)
    print(b.is_needed_to_heat(tmp_act, tmp_goal, time_to_consumption))



from pathlib import Path
print('Running' if __name__ == '__main__' else 'Importing', Path(__file__).resolve())
from datetime import timedelta
import datetime
from sklearn.preprocessing import RobustScaler
from tensorflow.keras.models import Sequential 
import tensorflow as tf

from keras.layers import LSTM
from keras.layers import Dropout
from keras.models import Model
from keras.layers import Input, Dense
from keras.models import load_model
from keras.callbacks import EarlyStopping, ModelCheckpoint
import keras.backend as K

import numpy as np
import pandas as pd
from scipy import stats
from smartboiler.data_handler import DataHandler

class Forecast:
    def __init__(self, dataHandler: DataHandler):
        self.batch_size = 3
        self.lookback = 10
        self.delay = 1
        self.predicted_column = 'longtime_mean'
        self.dataHandler = dataHandler
        self.scaler = RobustScaler()


    
    def train_model(self, begin_of_training = None, end_of_training = None):
        
        df, _ = self.dataHandler.get_data_for_training_model(left_time_interval=begin_of_training, right_time_interval=end_of_training, predicted_column=self.predicted_column)
        self.num_of_features = len(df.columns) - 1
        self.df_train_norm = df.copy()


        self.df_train_norm[df.columns] = self.scaler.fit_transform(df)

        self.train_gen = self.generator(dataframe = self.df_train_norm, 
                      target_name = self.predicted_column, 
                      lookback = self.lookback,
                      delay = self.delay,
                      min_index = 0,
                      max_index = int(df.shape[0]*0.8),
                      step = 1,
                      shuffle = True,
                      batch_size = self.batch_size)

        self.valid_gen = self.generator(dataframe = self.df_train_norm, 
                            target_name = self.predicted_column, 
                            lookback = self.lookback,
                            delay = self.delay,
                            min_index = int(df.shape[0]*0.8),
                            max_index = None,
                            step = 1,
                            shuffle = False,
                            batch_size = self.batch_size)
        

        
        self.val_steps = int((self.df_train_norm.shape[0]*0.1 - self.lookback) // self.batch_size)
        # This is how many steps to draw from `train_gen`
        # in order to see the whole train set:
        self.train_steps = int((self.df_train_norm.shape[0]*0.9 - self.lookback) // self.batch_size)


        
    def generator(self, dataframe, target_name, lookback, delay, min_index, max_index,shuffle=False, batch_size=128, step=6):
    
        data = dataframe.values
        data = data.astype(np.float32)
        target_indx = dataframe.columns.get_loc(target_name)
        
        if max_index is None:
            max_index = len(data) - delay - 1
        i = min_index + lookback
        while 1:
            if shuffle:
                rows = np.random.randint(
                    min_index + lookback, max_index, size=batch_size)
            else:
                if i + batch_size >= max_index:
                    i = min_index + lookback
                rows = np.arange(i, min(i + batch_size, max_index))
                i += len(rows)

            samples = np.zeros((len(rows),
                            lookback // step,
                            data.shape[-1]))
            targets = np.zeros((len(rows),))
            for j, row in enumerate(rows):
                indices = range(rows[j] - lookback, rows[j], step)
                samples[j] = data[indices]
                targets[j] = data[rows[j] + delay][target_indx]
            yield samples, targets
            
    def r2_keras(self, y_true, y_pred):
        """Coefficient of Determination 
        """
        SS_res =  K.sum(K.square( y_true - y_pred ))
        SS_tot = K.sum(K.square( y_true - K.mean(y_true) ) )
        return ( 1 - SS_res/(SS_tot + K.epsilon()) )
    
    def build_model(self):
        
        model = Sequential()
        model.add(tf.keras.Input(shape=(None, self.df_train_norm.shape[1])))
        model.add(tf.keras.layers.Conv1D(filters=6, kernel_size=5, activation='relu'))
        model.add(LSTM(6, return_sequences=True, activation='relu'))
        model.add(LSTM(6, return_sequences=False, activation='relu'))
        model.add(Dense(1))

        self.model = model
    
    def fit_model(self):
        callbacks = [EarlyStopping(monitor='loss', min_delta = 0, patience=10, verbose=1, mode='auto', restore_best_weights=True),
             ModelCheckpoint(filepath='lstm_model.h5', monitor='val_loss', save_best_only=True)]

        self.model.compile(loss='mae', optimizer='adam',metrics=[self.r2_keras])
        # history = model.fit(train_gen, epochs=50, batch_size=72, validation_data=valid_gen, verbose=2, shuffle=False, use_multiprocessing=True)

        history = self.model.fit(self.train_gen,
                                    steps_per_epoch=self.train_steps,
                                    epochs=70,
                                    shuffle=False,
                                    validation_data=self.valid_gen,
                                    validation_steps=self.val_steps,
                                    callbacks = callbacks)
        
    def get_forecast_next_steps(self, left_time_interval = None, right_time_interval = None, predicted_column = 'longtime_mean'):
        
        if (left_time_interval is None):
            left_time_interval = datetime.now() - timedelta(days=2)
        if (right_time_interval is None):
            right_time_interval = datetime.now()        
        
        df_all = self.dataHandler.get_data_for_prediction(left_time_interval=left_time_interval, right_time_interval=right_time_interval, predicted_column=self.predicted_column)

        forecast_future = pd.DataFrame()

        current_forecast_begin_date = left_time_interval
        current_forecast_end_date = left_time_interval + timedelta(minutes=30)
        for i in range(0, 48):
            print("-----")

            df_predict = pd.DataFrame({'datetime': pd.date_range(current_forecast_begin_date, current_forecast_end_date, freq='30min')})
            df_predict['longtime_mean'] = 0
            df_predict['weekday_sin'] = np.sin(2 * np.pi * df_predict['datetime'].dt.weekday / 7)
            df_predict['weekday_cos'] = np.cos(2 * np.pi * df_predict['datetime'].dt.weekday / 7)
            df_predict['hour_sin'] = np.sin(2 * np.pi * df_predict['datetime'].dt.hour / 24)
            df_predict['hour_cos'] = np.cos(2 * np.pi * df_predict['datetime'].dt.hour / 24)
            df_predict['minute_sin'] = np.sin(2 * np.pi * df_predict['datetime'].dt.minute / 60)
            df_predict['minute_cos'] = np.cos(2 * np.pi * df_predict['datetime'].dt.minute / 60)
            # delete column datetime
            df_predict = df_predict.drop(columns='datetime')
            len_df_predict = df_predict.shape[0]
            
            
            # concar df_all and df_predict
            
            df_all = pd.concat([df_all, df_predict], axis=0)
            df_all = df_all.reset_index(drop=True)
            
            df_predict_norm = df_all.copy()
            df_predict_norm[df_all.columns] = self.scaler.transform(df_all)
            # create predict df with values 

            predict_gen = self.generator(dataframe = df_predict_norm, 
                    target_name = self.predicted_column, 
                    lookback = self.lookback,
                    delay = self.delay,
                    min_index = 0,
                    max_index = None,
                    step = 1,
                    shuffle = False,
                    batch_size = df_predict.shape[0])

            (X, y_truth) = next(predict_gen)

            y_pred = self.model.predict(X)

            # np.expand_dims(y_truth,axis=1).shape
            y_pred_inv = np.concatenate((y_pred,np.zeros((y_pred.shape[0],self.num_of_features))),axis=1)
            y_pred_inv = self.scaler.inverse_transform(y_pred_inv)
            y_pred_inv = y_pred_inv[:,0]
            
            # set df_all last len_df_predict values to y_pred_inv
            df_all.iloc[-len_df_predict:, df_all.columns.get_loc(predicted_column)] = y_pred_inv
            # plt.plot(df_all.iloc[-len_df_predict:, df_all.columns.get_loc('longtime_mean')], color = 'green', label = 'Predicted data')
            df_all = df_all[len_df_predict:]
            forecast_future = pd.concat([forecast_future, df_all[-len_df_predict:]], axis=0)
            forecast_future = forecast_future.reset_index(drop=True)
            
            current_forecast_begin_date = current_forecast_begin_date + timedelta(hours=0.5)
            current_forecast_end_date = current_forecast_end_date + timedelta(hours=0.5)
        
        # create a dataframe with forecast and datetime as index
        return forecast_future
    


        
if __name__ == '__main__':
    dataHandler = DataHandler(
        "localhost",
        "smart_home_formankovi",
        "root",
        "root",
        "shelly1pm_34945475a969",
        "esphome_web_c771e8_tmp3"
    )    
    forecast = Forecast(dataHandler)
    forecast.train_model()
    forecast.build_model()
    forecast.fit_model()
    forecast.get_forecast_next_steps(30)
        
    


##########################################################
# Bachelor's thesis                                      #
# From a dumb boiler to a smart one using a smart socket #
# Author: Adam Grünwald                                  #
# BUT FIT BRNO, Faculty of Information Technology        #
# 26/6/2021                                              #
#                                                        #
# Module with class used for creating schedules          #
# of predicted consumptions                              #
##########################################################

from scipy.signal import argrelextrema, find_peaks, peak_widths
from datetime import datetime, timedelta, date
from time_handler import TimeHandler

import pandas as pd
import numpy as np


class WeekPlanner:

    def __init__(self, data):
        print("------------------------------------------------------\n")
        print('initializing of WeekPlanner\n')
        print("------------------------------------------------------\n")
        self.TimeHandler = TimeHandler()
        self.week_days_consumptions = self._create_days_average(data)
        print(self.week_days_consumptions)
        self.week_days_high_tarifs_intervals = self._find_empty_intervals(data)


        #creating of dictionary of day coefficients
        self.week_days_coefs = dict.fromkeys(range(0, 7))
        for key in self.week_days_coefs:
            self.week_days_coefs[key] = 1.3

    def _find_empty_intervals(self, data):
        """Finds an empty intervals in measured dataframe, 
        which represents the times of high tarif, when the boiler cannot heat.

        Args:
            data ([DataFrame]): [DataFrame to search in]

        Returns:
            [dict]: [times of high tarifs]
        """

        time_interval_int = 5
        time_interval = str(time_interval_int) + 'min'

        data = data[data.index > (
            data.last_valid_index() - timedelta(days=14))]

        # grouping data from db by time interval in minutes
        data = data.groupby(pd.Grouper(freq=time_interval)).aggregate(np.mean)

        # grouping data grouped by time interval by dayofweek, hour and minut
        df_grouped = data.groupby(
            [data.index.dayofweek, data.index.hour, data.index.minute], dropna=True).mean()

        # start and of measuring for creating an emptu dataframe with rows for all times by 5 minuts
        start = data.first_valid_index()
        end = data.last_valid_index()
        # new empty dataframe
        df = pd.DataFrame({'tmp2': -1}, index=pd.date_range(start,
                                                            end, freq='1min'))
        # grouped by same time interval as dataframe from db and then by day of week, hour and minute
        df = df.groupby(pd.Grouper(freq=time_interval)).aggregate(np.sum)
        df = df.groupby([df.index.dayofweek, df.index.hour,
                        df.index.minute], dropna=False).sum()

        # adding of values from db into an empty dataframe
        df.tmp2 = df_grouped.tmp2

        week_high_tarifs = {}

        # for each of day of week finds intervals of high tarif
        for idx in range(7):

            i = 0  # index of interval of the day

            day_high_tarifs = {}
            first_none = False #first empty interval
            for index, row in df.iloc[df.index.get_level_values(level=0) == idx].iterrows():
                #iterates over values in data from certain day of week

                day_of_week = index[0]
                hour = index[1]
                minute = index[2]

                tmp1 = row[0]

                if(np.isnan(tmp1)):
                    if (first_none == False):
                        #found first none value
                        first_none = True

                        start_of_none = timedelta(hours=(hour + (minute/60)))
                        minute -= time_interval_int

                        if minute < 0:
                            minute += time_interval_int
                            hour -= 1
                            if hour < 0:
                                hour += 1
                                day_of_week = (day_of_week+1) % 7
                        previous_index = (day_of_week, hour, minute)

                        tmp_before_start = df.loc[previous_index][0]

                else:
                    if first_none:
                        end_of_none = timedelta(hours=(hour + (minute/60)))
                        duration = (end_of_none - start_of_none)

                        end_of_none = (datetime.min + end_of_none).time()
                        start_of_none = (datetime.min + start_of_none).time()

                        minute += time_interval_int

                        if minute >= 60:
                            minute -= time_interval_int
                            hour += 1
                            if hour > 23:
                                hour -= 1
                                day_of_week = (day_of_week-1) % 7

                        next_index = (day_of_week, hour, minute)
                        tmp_after_end = df.loc[next_index][0]

                        tmp_delta = tmp_before_start - tmp_after_end
                        #saves empty interval to dict of empty intervals for one day
                        day_high_tarifs.update(
                            {i: {"start": start_of_none, "end": end_of_none, "duration": duration, "tmp_delta": tmp_delta}})
                        i += 1
                        first_none = False
            #updates dict of week empty intervals with day empty intervals
            week_high_tarifs.update({idx: day_high_tarifs})

        return week_high_tarifs

    def _create_days_average(self, data):
        """creating of new days averages with peaks for predicting consumption

        Args:
            data ([DataFrame]): [dataframe to create averages for prediction from]

        Returns:
            [dict]: [dictionary with predicted consumption for each day of week]
        """

        new_week_days_consumptions = dict.fromkeys(
            self.TimeHandler.daysofweek, 0)

        prepared_data = self._prepare_data(data)

        for idx in range(7):

            x = prepared_data[prepared_data.day_of_week ==
                              idx].tmp1.reset_index(drop=True)

            peaks, _ = self._find_ideal_peaks(x)

            results_half = peak_widths(x, peaks, rel_height=0.7)

            day_aconsumptions = {}
            for i in range(len(results_half[0])):
                day_aconsumptions.update({i: {"start": self.TimeHandler.float_to_time(results_half[2][i]),
                                              "end": self.TimeHandler.float_to_time(results_half[3][i]),
                                              "duration": results_half[0][i],
                                              "peak": round(_['peak_heights'][i], 2)
                                              }
                                          })

            new_week_days_consumptions.update({idx: day_aconsumptions})

        print("created new days average")

        return new_week_days_consumptions

    def _find_ideal_peaks(self, x):
        """searchs for maximal 4 peaks in day data

        Args:
            x ([DataFrame]): [dataframe for specific day of week]
        """
        height = 20
        distance = 3
        peaks, _ = find_peaks(x, height=height, distance=distance)

        while(len(peaks) > 4):
            height += 1
            distance += 0.2
            peaks, _ = find_peaks(x, height=height, distance=distance)

        return(peaks, _)

    def _prepare_data(self, data):
        """groups the measured data by day of week and 1 hour

        Args:
            data ([type]): [description]

        Returns:
            [type]: [description]
        """

        
        data = data[data.index > (
            data.last_valid_index() - timedelta(days=14))]
            
        df_grouped = data.groupby(
            [data.index.dayofweek, data.index.hour]).mean().reset_index()
        return df_grouped.rename(columns={'level_0': 'day_of_week', 'level_1': 'hour'})

    def week_plan(self, data_from_db=None):
        """Accepts data from DB and returns plan for week based on days of week average.
        If data wasn't specified, returns last created week plan.

        Args:
            data ([result of query from dB]): data from DB grouped by minute
        """

        if ((data_from_db is not None)):
            self.week_days_consumptions = self._create_days_average(
                data_from_db)
            self.week_days_high_tarifs_intervals = self._find_empty_intervals(
                data_from_db)

        print("created new week plan!")

        return self.week_days_consumptions


    def next_heating_event(self, event):
        """Finds how long it takes to next heating.

        Returns:
            [type]: [description]
        """
        actual_time = self.TimeHandler.hour_minutes_now()

        day_of_week = datetime.now().weekday()

        days_plus = 0

        while(days_plus < 7):

            day_plan = self.week_days_consumptions[day_of_week]

            for key, item in day_plan.items():
                next_time = item[event]

                if (next_time >= actual_time):
                    time_to_next_heating_event = (
                        next_time - actual_time + timedelta(days=days_plus)) / timedelta(hours=1)

                    return{"will_occur_in": time_to_next_heating_event, "duration": item['duration'], "peak": item["peak"], "time": next_time}

            actual_time = self.TimeHandler.hour_minutes_now().replace(hour=0, minute=0)
            days_plus += 1
            day_of_week = (day_of_week + 1) % 7
        return None

    def is_in_heating(self):
        end = self.next_heating_event('end')
        start = self.next_heating_event('start')

        if ((end is None) or (start is None)):
            return False
        hours_to_end = end["will_occur_in"]
        hours_to_start = start["will_occur_in"]

        return hours_to_start > hours_to_end

    def is_in_DTO(self):
        """Checks whether is DTO currently turned on

        Returns:
            [boolean]: [boolean value specifiing if it is in DTO]
        """

        start = self.next_high_tarif_interval('start')
        if start is not None:

            time_to_next_DTO_start = start['next_high_tarif_in']
            time_to_next_DTO_end = self.next_high_tarif_interval('end')[
                'next_high_tarif_in']

            return time_to_next_DTO_start > time_to_next_DTO_end

        return False

    def next_high_tarif_interval(self, event):
        """Returns the next high tarif interval with time to next interval 
        and delta of temperatures before and after

        Args:
            event ([string]): [type of event to search for]

        Returns:
            [dict]: [dict of next high tarif interval]
        """

        actual_time = datetime.now().time()

        day_of_week = datetime.now().weekday()
        days_plus = 0

        while(days_plus < 7):

            day_high_tarifs_intervals = self.week_days_high_tarifs_intervals[day_of_week]

            for key, item in day_high_tarifs_intervals.items():
                next_time = item[event]

                next_actual_time_delta = datetime.combine(
                    date.min, next_time) - datetime.combine(date.min, actual_time)

                if (next_actual_time_delta >= timedelta(minutes=0)):
                    time_to_next_heating_event = (
                        next_actual_time_delta + timedelta(days=days_plus)) / timedelta(hours=1)

                    return {"next_high_tarif_in": time_to_next_heating_event, "tmp_delta": item['tmp_delta']}

            actual_time = actual_time.replace(hour=0, minute=0)
            days_plus += 1
            day_of_week = (day_of_week + 1) % 7
        return None

    def duration_of_low_tarif_to_next_heating(self, hours_to_next_heating):
        """Returns duration of low tarif to next heating. It is time to next consumption
         substracted by time of high tarif in this interval. 

        Args:
            hours_to_next_heating ([float]): [hours remaining to next heating]

        Returns:
            [float]: [time to next consumption minus time of high interval]
        """
        datetime_now = datetime.now()
        time_now = datetime_now.time()
        day_time_start = time_now

        hours_to_next_heating = timedelta(hours=hours_to_next_heating)

        day_of_week = datetime_now.weekday()

        week_high_tarifs_intervals = self.week_days_high_tarifs_intervals

        time_added = timedelta(hours=0)
        time_of_high_tarif = timedelta(hours=0)

        while((time_added < hours_to_next_heating) and day_of_week < 7):
            day_high_tarifs_intervals = week_high_tarifs_intervals[day_of_week]
            if(day_high_tarifs_intervals is not None):

                added_in_day = timedelta(hours=0)

                for j in range(len(day_high_tarifs_intervals)):
                    actual_interval = day_high_tarifs_intervals[j]

                    if time_now > actual_interval['end']:
                        continue

                    added_in_day += datetime.combine(
                        date.min, actual_interval['start']) - datetime.combine(date.min, time_now)

                    if ((time_now < actual_interval['end']) and (time_now > actual_interval['start'])):
                        time_of_high_tarif += datetime.combine(
                            date.min, actual_interval['end']) - datetime.combine(date.min, time_now)

                    elif added_in_day + time_added < hours_to_next_heating:
                        time_of_high_tarif += actual_interval['duration']

                    time_now = actual_interval['end']

            time_added += datetime.combine(date.min, time_now.replace(
                hour=23, minute=59)) - datetime.combine(date.min, day_time_start)
            day_of_week = (day_of_week + 1) % 7
            time_now = time_now.replace(hour=0, minute=0)
            day_time_start = time_now

        return (hours_to_next_heating - time_of_high_tarif) / timedelta(hours=1)

    def _empty_intervals(self, data_from_db=None):
        """Function returning a dictionary of empty intervals.

        Args:
            data_from_db ([DataFrame], optional): [Data rom the DB to search in]. Defaults to None.

        Returns:
            [type]: [description]
        """
        if (data_from_db is not None):
            self.empty_intervals = self._find_empty_intervals(data_from_db)

        print('empty intervals found')

        return self.empty_intervals
